import logging

import rich_click as click

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.display_util import print_json
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import workspace as workspace_lib

logger = logging.getLogger(__name__)

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="delete", cls=GROUP_CLS)
def delete_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Servicefoundry delete resource

    \b
    Supported resources:
    - Workspace
    - Service
    """
    pass


@click.command(name="workspace", cls=COMMAND_CLS, help="Delete a Workspace")
@click.argument("name", type=click.STRING)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to delete the workspace from",
)
@click.option("--force", is_flag=True, default=False, help="force delete the workspace")
@click.confirmation_option(prompt="Are you sure you want to delete this workspace?")
@handle_exception_wrapper
def delete_workspace(name, cluster, force: bool = False):
    # Tests:
    # - Set Context -> delete workspace -> Should give error to give workspace name
    # - Set Context -> delete workspace valid_name -> Should delete
    # - Set Context -> delete workspace invalid_name -> Should give error no such workspace in set cluster
    # - Set Context -> delete workspace name -c 'invalid_cluster_name' -> Should give error invalid cluster
    # - Set Context -> delete workspace invalid_name -c 'cluster_name' -> Should give error invalid workspace
    # - Set Context -> delete workspace valid_name -c 'cluster_name' -> Should delete
    # - No Context -> delete workspace -> Should give error to give workspace name
    # - No Context -> delete workspace valid_name -> Try to resolve, if only one exists then delete
    #                 otherwise error to give cluster
    # - No Context -> delete workspace invalid_name -> Tries to resolve, if only one exists then delete
    #                 otherwise error to give cluster
    # - No Context -> delete workspace name -c 'invalid_cluster_name' -> Should give error invalid cluster
    # - No Context -> delete workspace invalid_name -c 'cluster_name' -> Should give error invalid workspace
    # - No Context -> delete workspace valid_name -c 'cluster_name' -> Should delete
    response = workspace_lib.delete_workspace(
        name_or_id=name,
        cluster_name_or_id=cluster,
        force=force,
        non_interactive=True,
    )
    if CliConfig.get("json"):
        print_json(data=response)


def get_delete_command():
    delete_command.add_command(delete_workspace)
    return delete_command



import warnings

from pydantic import Field, constr, root_validator

from servicefoundry.auto_gen import models
from servicefoundry.lib.model.entity import Deployment
from servicefoundry.logger import logger
from servicefoundry.v2.lib.deploy import deploy_component

_TOP_LEVEL_COMMAND_DEPRECATION_MESSAGE = """
==================== DEPRECATION WARNING ====================
`command` has been deprecated and will be removed in future.
Please use `command` argument of `Image` or `DockerFileBuild` or `PythonBuild` instead.
E.g.
    {cls_name}(..., image=Image(..., command=...))
    OR {cls_name}(..., image=Build(..., build_spec=DockerFileBuild(..., command=...)))
    OR {cls_name}(..., image=Build(..., build_spec=PythonBuild(..., command=...)))

These are equivalent to `image.command` and `image.build_spec.command in yaml spec

E.g.

```
name: "{name}"
image:
  type: "image"
  command: ...
```

OR

```
name: "{name}"
image:
  type: "build"
  build_spec:
    type: "tfy-python-buildpack" # or "dockerfile"
    command: ...
```

============================================================
"""


def _warn_top_level_command_use(cls, values):
    command = values.get("command")
    image = values.get("image")
    if command and image:
        _message = _TOP_LEVEL_COMMAND_DEPRECATION_MESSAGE.format(
            cls_name=cls.__name__, name=values.get("name") or "NAME_NOT_FOUND"
        )
        logger.warning(_message)
        warnings.warn(_message, category=DeprecationWarning, stacklevel=2)
        if image.type == "image":
            values["image"].command = command
        elif image.type == "build":
            values["image"].build_spec.command = command
        values.pop("command")
    return values


class Service(models.Service):
    type: constr(regex=r"service") = "service"
    resources: models.Resources = Field(default_factory=models.Resources)

    class Config:
        extra = "forbid"

    _patch_command = root_validator(allow_reuse=True)(_warn_top_level_command_use)

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn, wait=wait)


class Job(models.Job):
    type: constr(regex=r"job") = "job"
    resources: models.Resources = Field(default_factory=models.Resources)

    class Config:
        extra = "forbid"

    _patch_command = root_validator(allow_reuse=True)(_warn_top_level_command_use)

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn, wait=wait)


class Notebook(models.Notebook):
    type: constr(regex=r"notebook") = "notebook"
    resources: models.Resources = Field(default_factory=models.Resources)

    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn)


class ModelDeployment(models.ModelDeployment):
    type: constr(regex=r"model-deployment") = "model-deployment"
    resources: models.Resources = Field(default_factory=models.Resources)

    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(component=self, workspace_fqn=workspace_fqn, wait=wait)


class Application(models.Application):
    class Config:
        extra = "forbid"

    def deploy(self, workspace_fqn: str, wait: bool = True) -> Deployment:
        return deploy_component(
            component=self.__root__, workspace_fqn=workspace_fqn, wait=wait
        )



from __future__ import annotations

import datetime
import time
from enum import Enum
from typing import Any, ClassVar, Dict, List, Optional

import jwt
from pydantic import BaseModel, Extra, Field, constr, validator

# TODO: switch to Enums for str literals
# TODO: Need a better approach to keep fields in sync with server
#       most fields should have a default in case server adds/removes a field
# TODO: Implement NotImplementedError sections


class Base(BaseModel):
    class Config:
        validate_assignment = True
        use_enum_values = True
        extra = Extra.allow

    def __repr_args__(self):
        return [
            (key, value)
            for key, value in self.__dict__.items()
            if self.__fields__[key].field_info.extra.get("repr", True)
        ]

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]):
        return cls.parse_obj(dct)

    def to_dict(self) -> Dict[str, Any]:
        return self.dict()


class Entity(Base):
    createdAt: datetime.datetime = Field(repr=False)
    updatedAt: datetime.datetime = Field(repr=False)
    list_display_columns: ClassVar[List[str]] = []
    get_display_columns: ClassVar[List[str]] = []


class Cluster(Entity):
    id: str = Field(repr=False)
    name: str
    fqn: str
    region: str = Field(repr=False)
    isTenantDefault: bool = False
    list_display_columns: ClassVar[List[str]] = [
        "name",
        "fqn",
        "region",
        "createdAt",
    ]
    get_display_columns: ClassVar[List[str]] = [
        "name",
        "fqn",
        "region",
        "createdAt",
        "updatedAt",
    ]

    def to_dict_for_session(self) -> Dict[str, Any]:
        return self.dict()

    @property
    def workspaces(self) -> List["Workspace"]:
        raise NotImplementedError


class Workspace(Entity):
    id: str = Field(repr=False)
    fqn: str
    name: str
    clusterId: str = Field(repr=False)
    createdBy: str = Field(repr=False)
    list_display_columns: ClassVar[List[str]] = [
        "name",
        "fqn",
        "createdAt",
    ]
    get_display_columns: ClassVar[List[str]] = [
        "name",
        "fqn",
        "createdBy",
        "createdAt",
        "updatedAt",
    ]

    def to_dict_for_session(self) -> Dict[str, Any]:
        return self.dict()

    @property
    def cluster(self) -> Cluster:
        raise NotImplementedError


class NewDeployment(Entity):
    def __init__(self, **data: Any):
        super().__init__(**data)
        self.workspaceName = self.workspace["name"]
        self.clusterName = (
            self.workspace.get("metadata", {}).get("manifest", {}).get("cluster", "")
        )
        self.workspaceDetails = f'{self.workspace["name"]} ({self.clusterName})'
        self.type = self.manifest["type"]

    id: str = Field(repr=False)
    version: str = Field(repr=False)
    fqn: str
    applicationId: str = Field(repr=False)
    workspaceId: str = Field(repr=False)

    # metadata: Optional[Dict[str, Any]] = Field(repr=False)
    manifest: Dict[str, Any] = Field(repr=False)

    failureReason: Optional[str] = Field(repr=False)
    application: Optional[Dict[str, Any]] = Field(repr=False)
    workspace: Dict[str, Any] = Field(repr=False)
    baseDomainURL: str = Field(repr=False)
    createdBy: str = Field(repr=False)
    workspaceName: Optional[str] = Field(repr=False)
    clusterName: Optional[str] = Field(repr=False)
    workspaceDetails: Optional[str] = Field(repr=False)
    type: Optional[str] = Field(repr=False)

    list_display_columns: ClassVar[List[str]] = [
        "fqn",
        "workspaceDetails",
        "type",
        "createdAt",
    ]


# TODO: Should treat displaying and handling these with more respect as it is sensitive data


class WorkspaceResources(BaseModel):
    cpu_limit: Optional[float]
    memory_limit: Optional[int]
    ephemeral_storage_limit: Optional[int]


class UserType(str, Enum):
    USER: str = "user"
    GROUP: str = "group"


class User(BaseModel):
    name: str
    type: UserType


class RoleBinding(BaseModel):
    workspace_admin: List[User] = Field(alias="workspace-admin")
    workspace_editor: List[User] = Field(alias="workspace-editor")
    workspace_viewer: List[User] = Field(alias="workspace-viewer")

    class Config:
        allow_population_by_field_name = True


class Deployment(BaseModel):
    id: str
    fqn: str
    version: int
    # TODO: Dict -> pydantic model if required
    manifest: Dict[str, Any]
    # workspace: Dict[str, Any]
    # TODO: make status an enum
    createdBy: str
    applicationId: str
    failureReason: Optional[str]
    createdAt: datetime.datetime
    updatedAt: datetime.datetime
    # TODO: Dict -> pydantic model if required
    # application: Dict[str, Any]
    # TODO: Dict -> pydantic model if required
    # workspace: Dict[str, Any]
    # baseDomainURL: str
    # builds: List[BuildResponse]

    class Config:
        extra = Extra.allow


class PortMetadata(BaseModel):
    port: int
    host: str


class DeploymentMetadata(BaseModel):
    name: str
    ports: List[PortMetadata]


class DeploymentTransitionStatus(str, Enum):
    INITIALIZED: str = "INITIALIZED"
    BUILDING: str = "BUILDING"
    DEPLOYING: str = "DEPLOYING"
    BUILD_SUCCESS: str = "BUILD_SUCCESS"
    DEPLOY_SUCCESS: str = "DEPLOY_SUCCESS"
    DEPLOY_FAILED: str = "DEPLOY_FAILED"
    BUILD_FAILED: str = "BUILD_FAILED"
    CANCELLED: str = "CANCELLED"
    FAILED: str = "FAILED"
    _: str = ""

    @classmethod
    def is_failure_state(cls, state: DeploymentTransitionStatus) -> bool:
        return state in (cls.BUILD_FAILED, cls.FAILED)


class DeploymentState(BaseModel):
    isTerminalState: bool


class DeploymentStatus(BaseModel):
    state: DeploymentState
    status: DeploymentTransitionStatus
    transition: Optional[DeploymentTransitionStatus] = None


class DeploymentInfo(Deployment):
    metadata: Optional[List[DeploymentMetadata]]
    currentStatus: DeploymentStatus
    manifest: Dict[str, Any]


class ApplicationInfo(BaseModel):
    lastVersion: int
    activeVersion: int
    deployment: DeploymentInfo
    workspace: Dict[str, Any]
    fqn: str


class UserInfo(BaseModel):
    user_id: constr(min_length=1)
    email: str
    tenant_name: constr(min_length=1) = Field(alias="tenantName")

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False


class TenantInfo(BaseModel):
    tenant_name: constr(min_length=1) = Field(alias="tenantName")
    auth_server_url: str

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False


class Token(BaseModel):
    access_token: constr(min_length=1) = Field(alias="accessToken", repr=False)
    refresh_token: constr(min_length=1) = Field(alias="refreshToken", repr=False)
    decoded_value: Optional[Dict] = Field(exclude=True, repr=False)

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    @validator("decoded_value", always=True, pre=True)
    def _decode_jwt(cls, v, values, **kwargs):
        access_token = values["access_token"]
        return jwt.decode(
            access_token,
            options={
                "verify_signature": False,
                "verify_aud": False,
                "verify_exp": False,
            },
        )

    @property
    def tenant_name(self) -> str:
        return self.decoded_value["tenantName"]

    def is_going_to_be_expired(self, buffer_in_seconds: int = 120) -> bool:
        exp = int(self.decoded_value["exp"])
        return (exp - time.time()) < buffer_in_seconds

    def to_user_info(self) -> UserInfo:
        return UserInfo(
            user_id=self.decoded_value["username"],
            email=self.decoded_value["email"],
            tenant_name=self.tenant_name,
        )


class CredentialsFileContent(BaseModel):
    access_token: constr(min_length=1) = Field(repr=False)
    refresh_token: constr(min_length=1) = Field(repr=False)
    host: constr(min_length=1)

    class Config:
        allow_mutation = False

    def to_token(self) -> Token:
        return Token(access_token=self.access_token, refresh_token=self.refresh_token)


class DeviceCode(BaseModel):
    user_code: str = Field(alias="userCode")
    device_code: str = Field(alias="deviceCode")

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    def get_user_clickable_url(self, auth_host: str) -> str:
        return f"{auth_host}/authorize/device?userCode={self.user_code}"



# NOTE: Later all of these can be under a `models` or some other module
# I am not entirely sure about the structure of `lib` module yet
# This can go through another round of refactoring
from servicefoundry.v2.lib.deployble_patched_models import (
    Application,
    Job,
    ModelDeployment,
    Notebook,
    Service,
)
from servicefoundry.v2.lib.patched_models import (
    BasicAuthCreds,
    Build,
    DockerFileBuild,
    FileMount,
    GitSource,
    HealthProbe,
    HttpProbe,
    HuggingfaceModelHub,
    Image,
    LocalSource,
    Manual,
    Param,
    Port,
    PythonBuild,
    RemoteSource,
    Resources,
    Schedule,
    TruefoundryModelRegistry,
)



import logging

import rich_click as click

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.console import console
from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.display_util import print_json, print_list
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.dao import deployment as deployment_lib
from servicefoundry.lib.dao import workspace as workspace_lib
from servicefoundry.lib.messages import PROMPT_NO_DEPLOYMENTS, PROMPT_NO_WORKSPACES
from servicefoundry.lib.model.entity import Cluster, NewDeployment, Workspace

logger = logging.getLogger(__name__)

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="list", cls=GROUP_CLS)
def list_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Servicefoundry list resources
    """


@click.command(name="cluster", cls=COMMAND_CLS, help="List Clusters")
@handle_exception_wrapper
def list_cluster():
    tfs_client = ServiceFoundryServiceClient()
    clusters = tfs_client.list_cluster()
    if CliConfig.get("json"):
        print_json(data=clusters)
    else:
        print_list("Clusters", clusters, Cluster.list_display_columns)


@click.command(name="workspace", cls=COMMAND_CLS, help="List Workspaces")
@click.option("-A", "--all", is_flag=True, default=False)
@click.option("-c", "--cluster", type=click.STRING, default=None, help="cluster name")
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def list_workspace(all, cluster, non_interactive):
    # Tests:
    # - Set Context -> list workspace -> Should get workspaces in set cluster
    # - Set Context -> list workspace -c 'cluster_name' -> Should get workspaces in given cluster
    # - Set Context -> list workspace -c 'invalid_cluster_name' -> Should give error invalid cluster
    # - Set Context -> list workspace -A -> Should give all workspaces across all clusters
    # - No Context -> list workspace -c 'cluster_name' -> Should get workspaces in given cluster
    # - No Context -> list workspace -c 'invalid_cluster_name' -> Should give error invalid cluster
    # - No Context -> list workspace -A -> Should give all workspaces across all clusters
    # ? No Context -> list workspace -> Should list workspaces if there is only cluster or ask for cluster name
    workspaces = workspace_lib.list_workspaces(
        cluster_name_or_id=cluster,
        all_=all,
        non_interactive=non_interactive,
    )
    if not workspaces:
        console.print(PROMPT_NO_WORKSPACES)
    else:
        workspaces.sort(key=lambda w: (w.fqn, w.createdAt))
    # TODO (chiragjn): Display columns here need to show cluster name!
    workspaces = [w.to_dict() for w in workspaces]
    if CliConfig.get("json"):
        print_json(data=workspaces)
    else:
        print_list(
            "Workspaces",
            workspaces,
            columns=Workspace.list_display_columns,
        )


@click.command(name="deployment", cls=COMMAND_CLS, help="List Deployments")
@click.option(
    "-w", "--workspace", type=click.STRING, default=None, help="workspace fqn"
)
@handle_exception_wrapper
def list_deployments(workspace):
    deployments = deployment_lib.list_deployments(
        workspace_fqn=workspace,
    )
    if not deployments:
        console.print(PROMPT_NO_DEPLOYMENTS)
    else:
        deployments.sort(key=lambda s: (s.fqn))

    deployments = [s.to_dict() for s in deployments]
    if CliConfig.get("json"):
        print_json(data=deployments)
    else:
        # TODO (chiragjn): Display columns here need to show workspace and cluster name!
        print_list(
            "Deployments",
            deployments,
            columns=NewDeployment.list_display_columns,
        )


def get_list_command():
    # list_command.add_command(list_workspace)
    list_command.add_command(list_deployments)

    # if ENABLE_CLUSTER_COMMANDS:
    #     list_command.add_command(list_cluster)

    return list_command



# https://docs.python.org/3/howto/logging.html#library-config
import logging

from rich.logging import RichHandler

logger = logging.getLogger("servicefoundry")


def add_cli_handler(level: int = logging.INFO, show_path=False):
    handler = RichHandler(level=level, show_path=show_path)
    handler.setLevel(level)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)



from typing import Union

from pydantic import Field, constr

from servicefoundry.auto_gen import models


class DockerFileBuild(models.DockerFileBuild):
    class Config:
        extra = "forbid"

    type: constr(regex=r"dockerfile") = "dockerfile"


class PythonBuild(models.PythonBuild):
    class Config:
        extra = "forbid"

    type: constr(regex=r"tfy-python-buildpack") = "tfy-python-buildpack"


class RemoteSource(models.RemoteSource):
    class Config:
        extra = "forbid"

    type: constr(regex=r"remote") = "remote"


class LocalSource(models.LocalSource):
    class Config:
        extra = "forbid"

    type: constr(regex=r"local") = "local"


class Build(models.Build):
    class Config:
        extra = "forbid"

    type: constr(regex=r"build") = "build"
    build_source: Union[
        models.RemoteSource, models.GitSource, models.LocalSource
    ] = Field(default_factory=LocalSource)


class Manual(models.Manual):
    class Config:
        extra = "forbid"

    type: constr(regex=r"manual") = "manual"


class Schedule(models.Schedule):
    class Config:
        extra = "forbid"

    type: constr(regex=r"scheduled") = "scheduled"


class GitSource(models.GitSource):
    class Config:
        extra = "forbid"

    type: constr(regex=r"git") = "git"


class HttpProbe(models.HttpProbe):
    class Config:
        extra = "forbid"

    type: constr(regex=r"http") = "http"


class BasicAuthCreds(models.BasicAuthCreds):
    class Config:
        extra = "forbid"

    type: constr(regex=r"basic_auth") = "basic_auth"


class TruefoundryModelRegistry(models.TruefoundryModelRegistry):
    class Config:
        extra = "forbid"

    type: constr(regex=r"tfy-model-registry") = "tfy-model-registry"


class HuggingfaceModelHub(models.HuggingfaceModelHub):
    class Config:
        extra = "forbid"

    type: constr(regex=r"hf-model-hub") = "hf-model-hub"


class HealthProbe(models.HealthProbe):
    class Config:
        extra = "forbid"


class Image(models.Image):
    class Config:
        extra = "forbid"

    type: constr(regex=r"image") = "image"


class Port(models.Port):
    class Config:
        extra = "forbid"


class Resources(models.Resources):
    class Config:
        extra = "forbid"


class Param(models.Param):
    class Config:
        extra = "forbid"


class FileMount(models.FileMount):
    class Config:
        extra = "forbid"



from typing import Optional

from servicefoundry.cli.console import console
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.messages import PROMPT_USING_WORKSPACE_CONTEXT
from servicefoundry.lib.model.entity import NewDeployment


def list_deployments(
    workspace_fqn: Optional[str] = None,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    # NOTES: this is horrible, needs refactoring
    from servicefoundry.lib.dao.workspace import get_workspace_by_fqn

    client = client or ServiceFoundryServiceClient()
    if not workspace_fqn:
        deployments = client.list_deployments()
    else:
        workspace = get_workspace_by_fqn(workspace_fqn)
        console.print(PROMPT_USING_WORKSPACE_CONTEXT.format(workspace.name))
        deployments = client.list_deployments(workspace.id)
    deployments = [NewDeployment.from_dict(d) for d in deployments]
    return deployments



import logging

import rich_click as click

from servicefoundry.cli.console import console
from servicefoundry.cli.const import COMMAND_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.io.rich_output_callback import RichOutputCallBack
from servicefoundry.lib.messages import PROMPT_POST_LOGIN
from servicefoundry.lib.session import login

logger = logging.getLogger(__name__)


@click.command(name="login", cls=COMMAND_CLS)
@click.option("--relogin", type=bool, is_flag=True, default=False)
@click.option("--host", type=str, default=None)
@click.option(
    "--api-key",
    type=str,
    default=None,
    prompt=True,
    hide_input=True,
    prompt_required=False,
)
@handle_exception_wrapper
def login_command(relogin: bool, host: str, api_key: str):
    """
    Login to servicefoundry
    """
    # TODO (chiragjn): Add support for non interactive login with API key.
    #                  It is supported indirectly as we always look for `SERVICE_FOUNDRY_API_KEY`
    #                  in the environment
    callback = RichOutputCallBack()
    login(api_key=api_key, host=host, relogin=relogin, output_hook=callback)
    console.print(PROMPT_POST_LOGIN)


def get_login_command():
    return login_command



import logging

import rich_click as click

from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.display_util import print_obj
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.dao import workspace as workspace_lib

logger = logging.getLogger(__name__)

WORKSPACE_DISPLAY_FIELDS = [
    "id",
    "name",
    "namespace",
    "status",
    "clusterId",
    "createdBy",
    "createdAt",
    "updatedAt",
]
DEPLOYMENT_DISPLAY_FIELDS = [
    "id",
    "serviceId",
    "domain",
    "deployedBy",
    "createdAt",
    "updatedAt",
]


@click.group(name="create", cls=GROUP_CLS)
def create_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Create servicefoundry resources

    \b
    Supported resources:
    - Workspace
    """
    pass


@click.command(name="workspace", cls=COMMAND_CLS, help="Create a new Workspace")
@click.argument("name", type=click.STRING)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to create this workspace in",
)
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def create_workspace(name, cluster, non_interactive):
    workspace = workspace_lib.create_workspace(
        name=name,
        cluster_name_or_id=cluster,
        non_interactive=non_interactive,
    )
    print_obj("Workspace", workspace.to_dict())


def get_create_command():
    create_command.add_command(create_workspace)
    return create_command



import json
import logging

import rich_click as click

from servicefoundry import builder
from servicefoundry.cli.const import GROUP_CLS
from servicefoundry.cli.util import handle_exception_wrapper

logger = logging.getLogger(__name__)


@click.group(
    name="build",
    cls=GROUP_CLS,
    invoke_without_command=True,
    help="Build servicefoundry Service",
)
@click.option("--name", type=click.STRING, default=None)
@click.option("--cache", type=click.STRING, default=None)
@click.option("--build-config", type=click.STRING, default=None)
@handle_exception_wrapper
def build_command(name, cache, build_config):
    if build_config:
        builder.build(
            build_configuration=json.loads(build_config), tag=name, cache_from=cache
        )
        return


def get_build_command():
    return build_command



import time

from model import Model
from model_composition import ModelComposition
from utils import preprocess

from servicefoundry.function_service import BuildConfig, FunctionService, remote

# composed_model = ModelComposition(model_1_path="foo", model_2_path="bar", model_3_path="baz")
composed_model = remote(
    ModelComposition,
    init_kwargs=dict(model_1_path="foo", model_2_path="bar", model_3_path="baz"),
)


preprocess_service = FunctionService(
    name="preprocess-service",
    build_config=BuildConfig(pip_packages=["numpy<1.22.0"]),
    port=7000,
)

preprocess_service.register_function(preprocess)


model_service = FunctionService(name="model-service", port=7001)


model_1 = remote(
    Model,
    init_kwargs={"model_path": composed_model.init_kwargs["model_1_path"]},
    name="model_1",
)
model_2 = remote(
    Model,
    init_kwargs={"model_path": composed_model.init_kwargs["model_2_path"]},
    name="model_2",
)
model_3 = remote(
    Model,
    init_kwargs={"model_path": composed_model.init_kwargs["model_2_path"]},
    name="model_3",
)
model_service.register_class(model_1)
model_service.register_class(model_2)
model_service.register_class(model_3)


# composed_model = remote(
#     ModelComposition,
#     init_kwargs=dict(model_1=model_1, model_2=model_2, model_3=model_3,),
# )
composed_model_service = FunctionService(name="composed-model-service", port=7002)
composed_model_service.register_class(composed_model)


print(preprocess_service)
print(model_service)
print(composed_model_service)

preprocess_service.run()
model_service.run()
composed_model_service.run()


time.sleep(1)

print(composed_model.predict(image_url="foo"))
time.sleep(1)


# 1.Multiple classes, each class can have its own route prefix.
# # 2.module vs filepath. Can I import a path directly as a module. This is relative to source.
# 3.What is the FQN for a normal function and a method. which I can detirministically generate at run time.
# 4. what if someone uses the same class and wnat to init it with two diff model path.



from typing import Sequence

import rich_click as click

from servicefoundry.cli.console import console
from servicefoundry.cli.const import COMMAND_CLS, GROUP_CLS
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.internal import experimental


@click.group(name="trigger", cls=GROUP_CLS)
def trigger_command():
    """
    Trigger a deployed job asynchronously
    """
    pass


@click.command(
    name="job",
    cls=COMMAND_CLS,
    context_settings=dict(ignore_unknown_options=True),
    help="Trigger a Job asynchronously",
)
@click.option(
    "--deployment_fqn",
    type=click.STRING,
    required=True,
    help="FQN of the deployment of the Job. This can be found on the Job details page.",
)
@click.argument("job_command", nargs=-1, type=click.UNPROCESSED)
@handle_exception_wrapper
def trigger_job(deployment_fqn: str, job_command: Sequence[str]):
    response = experimental.trigger_job(
        deployment_fqn=deployment_fqn, command=job_command
    )
    if job_command:
        message = f"Job triggered with command {job_command!r}"
    else:
        message = "Job triggered with pre-configured command"
    console.print(
        f"{message}.\n"
        f"It will be scheduled and visible at {response['previous_runs_url']}"
    )


def get_trigger_command():
    trigger_command.add_command(trigger_job)
    return trigger_command



import json

INVOKE_OP_CODE = "invoke:"
RESPONSE_OP_CODE = "response:"
ERROR_OP_CODE = "error:"
ACK_OP_CODE = "ack:"
DEPENDENCIES_OP_CODE = "dependency:"
FUNCTION_OP_CODE = "function:"
CLOSE_OP_CODE = "close:"


class Invoke:
    def __init__(self, function, payload):
        self.function = function
        self.payload = payload


class Response:
    def __init__(self, output):
        self.output = output


class ErrorResponse:
    def __init__(self, message, traceback):
        self.message = message
        self.traceback = traceback


class Close(ErrorResponse):
    pass


class Ack:
    pass


class Dependencies:
    pass


class Function:
    pass


class InvokeException(Exception):
    def __init__(self, message, traceback):
        super(InvokeException, self).__init__()
        self.message = message
        self.traceback = traceback

    def __str__(self):
        return self.message


class CloseException(Exception):
    def __init__(self, message, traceback):
        super(CloseException, self).__init__()
        self.message = message
        self.traceback = traceback

    def __str__(self):
        return self.message


class Communication:
    def __init__(self, infile, outfile):
        self.infile = infile
        self.outfile = outfile

    def read_input_msg(self):
        line = self.infile.readline()
        if line.startswith(INVOKE_OP_CODE):
            payload = json.loads(line[len(INVOKE_OP_CODE) :])
            return Invoke(function=payload["function"], payload=payload["payload"])
        elif line.startswith(RESPONSE_OP_CODE):
            payload = json.loads(line[len(RESPONSE_OP_CODE) :])
            return Response(output=payload["response"])
        elif line.startswith(ERROR_OP_CODE):
            payload = json.loads(line[len(ERROR_OP_CODE) :])
            raise InvokeException(payload["msg"], payload["traceback"])
        elif line.startswith(DEPENDENCIES_OP_CODE):
            return Dependencies()
        elif line.startswith(ACK_OP_CODE):
            return Ack()
        elif line.startswith(FUNCTION_OP_CODE):
            return Function()
        elif line.startswith(CLOSE_OP_CODE):
            payload = json.loads(line[len(CLOSE_OP_CODE) :])
            raise CloseException(payload["msg"], payload["traceback"])
        else:
            self.write_exception(f"Incorrect op code.")

    def read_response(self):
        response = self.read_input_msg()
        if isinstance(response, Response):
            return response.output
        raise RuntimeError(
            f"Unexpected response type {response.__class__}. Was expecting response or error."
        )

    def read_ack(self):
        response = self.read_input_msg()
        if isinstance(response, Ack):
            return
        raise RuntimeError(
            f"Unexpected response type {response.__class__}. Was expecting ack."
        )

    def write_ack(self):
        print(f"{ACK_OP_CODE}", file=self.outfile)

    def write_function(self):
        print(f"{FUNCTION_OP_CODE}", file=self.outfile)

    def write_invoke(self, function_name, **kwargs):
        invoke = {"function": function_name, "payload": kwargs}
        print(f"{INVOKE_OP_CODE}{json.dumps(invoke)}", file=self.outfile)

    def write_dependencies(self):
        print(f"{DEPENDENCIES_OP_CODE}", file=self.outfile)

    def write_response(self, res):
        res = {"response": res}
        print(f"{RESPONSE_OP_CODE}{json.dumps(res)}", file=self.outfile)

    def write_exception(self, msg, traceback=None):
        res = {"msg": msg, "traceback": traceback}
        print(f"{ERROR_OP_CODE}{json.dumps(res)}", file=self.outfile)

    def write_close(self, msg, traceback=None):
        res = {"msg": msg, "traceback": traceback}
        print(f"{CLOSE_OP_CODE}{json.dumps(res)}", file=self.outfile)

    def close(self):
        self.infile.close()
        self.outfile.close()



# This is meant to hold experimental features - stuff that is provisional and allowed to break
# DO NOT import stuff from here globally. Always import it locally restricted to as smaller scope as possible
# Always guard the imports under servicefoundry.lib.util.is_experimental_env_set
import functools
from typing import Any, Dict, Sequence, Union

from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.logger import logger


def _warn_on_call(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        logger.warning(
            f"Warning: This feature {fn.__name__} is in experimental stage. "
            f"As such there is no guarantees this will be maintained with backward compatibility "
            f"or even available moving forward"
        )
        return fn(*args, **kwargs)

    return wrapper


@_warn_on_call
def trigger_job(
    deployment_fqn: str, command: Union[str, Sequence[str]]
) -> Dict[str, Any]:
    client = ServiceFoundryServiceClient()
    deployment_info = client.get_deployment_info_by_fqn(deployment_fqn=deployment_fqn)
    deployment = client.get_deployment(
        application_id=deployment_info.applicationId,
        deployment_id=deployment_info.deploymentId,
    )
    application_info = client.get_application_info(
        application_id=deployment.applicationId
    )
    if deployment.version != application_info.activeVersion:
        raise Exception(
            f"The given `deployment_fqn` ({deployment_fqn!r}) belongs to an older version ({deployment.version}) "
            f"of the job. The latest active version is {application_info.activeVersion}. Triggering an older "
            f"version of the job is not supported. If you wish to run an older version, "
            f"you can redeploy the older version of the job first and then trigger that."
        )
    command_str = ""
    if command:
        if not isinstance(command, str):
            command_str = " ".join(command).strip()
        else:
            command_str = command.strip()
    command = command_str if command_str else None
    response = client._trigger_job(
        deployment_id=deployment_info.deploymentId,
        component_name=deployment.manifest["name"],
        command=command,
    )
    previous_runs_url = f"{client.base_url.strip('/')}/deployments/{deployment_info.applicationId}?tab=previousRuns"
    return {
        "status": response.get("message", "UNKNOWN"),
        "deployment_info": deployment_info,
        "previous_runs_url": previous_runs_url,
    }



import os
from typing import List, Optional, Tuple, TypeVar, Union

import questionary

from servicefoundry.cli.console import console
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.const import (
    SFY_DEBUG_ENV_KEY,
    SFY_EXPERIMENTAL_ENV_KEY,
    SFY_INTERNAL_ENV_KEY,
)
from servicefoundry.lib.messages import PROMPT_USING_CLUSTER_CONTEXT
from servicefoundry.lib.model.entity import Cluster, Workspace

# TODO: Move type casting downwards into `ServiceFoundrySession` and `ServiceFoundryServiceClient`
# TODO: Abstract duplicated code across resolving different entities
T = TypeVar("T")


def is_debug_env_set() -> bool:
    return True if os.getenv(SFY_DEBUG_ENV_KEY) else False


def is_experimental_env_set() -> bool:
    # TODO (chiragjn): one of these need to be removed
    return (
        True
        if os.getenv(SFY_EXPERIMENTAL_ENV_KEY) or os.getenv(SFY_INTERNAL_ENV_KEY)
        else False
    )


def _filter_by_attr_match(
    instances: List[T],
    value: Optional[str],
    attrs: Tuple[str, ...] = ("name", "fqn", "id"),
) -> List[T]:
    found = instances
    if value:
        for attr in attrs:
            found = [i for i in instances if getattr(i, attr, None) == value]
            if found:
                break
    return found


def all_clusters(client: ServiceFoundryServiceClient) -> List[Cluster]:
    clusters = [Cluster.from_dict(c) for c in client.list_cluster()]
    return clusters


def all_workspaces(client: ServiceFoundryServiceClient) -> List[Workspace]:
    workspaces = [Workspace.from_dict(w) for w in client.list_workspace()]
    return workspaces


def get_cluster_from_context(client: ServiceFoundryServiceClient) -> Optional[Cluster]:
    cluster = client.session.get_cluster()
    if cluster:
        cluster = Cluster.from_dict(cluster)
    return cluster


def get_workspace_from_context(
    client: ServiceFoundryServiceClient,
) -> Optional[Workspace]:
    workspace = client.session.get_workspace()
    if workspace:
        workspace = Workspace.from_dict(workspace)
    return workspace


def resolve_clusters(
    client: ServiceFoundryServiceClient,
    name_or_id: Optional[str] = None,
    ignore_context: bool = False,
) -> List[Cluster]:
    if not ignore_context and not name_or_id:
        cluster = get_cluster_from_context(client=client)
        if cluster:
            return [cluster]

    clusters = all_clusters(client=client)
    return _filter_by_attr_match(
        instances=clusters, value=name_or_id, attrs=("name", "fqn", "id")
    )


def resolve_workspaces(
    client: ServiceFoundryServiceClient,
    name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[Union[Cluster, str]] = None,
    ignore_context: bool = False,
) -> List[Workspace]:
    if not ignore_context and not name_or_id:
        workspace = get_workspace_from_context(client=client)
        if workspace:
            return [workspace]

    if isinstance(cluster_name_or_id, Cluster):
        clusters = [cluster_name_or_id]
        cluster_name_or_id = clusters[0].id
    else:
        clusters = resolve_clusters(
            client=client, name_or_id=cluster_name_or_id, ignore_context=ignore_context
        )

    if not clusters:
        if cluster_name_or_id:
            raise ValueError(f"No cluster found with name or id {cluster_name_or_id!r}")
        else:
            raise ValueError("Unable to resolve clusters without a name or id")
    elif len(clusters) == 1:
        cluster = clusters[0]
        _workspaces = client.get_workspace_by_name(
            workspace_name="", cluster_id=cluster.id
        )
        workspaces = [Workspace.from_dict(w) for w in _workspaces]
    else:
        if cluster_name_or_id:
            raise ValueError(
                f"More than one cluster found with name or id {cluster_name_or_id!r}: {clusters!r}"
            )
        else:
            # TODO (chiragjn): optimize this, we have wasted a call to v1/cluster
            workspaces = all_workspaces(client=client)

    return _filter_by_attr_match(
        instances=workspaces, value=name_or_id, attrs=("name", "fqn", "id")
    )


def ask_pick_cluster(clusters: List[Cluster]) -> Cluster:
    choices = [
        questionary.Choice(title=f"{c.name} ({c.fqn})", value=c) for c in clusters
    ]
    return questionary.select("Pick a cluster", choices=choices).ask()


def maybe_ask_pick_cluster(clusters: List[Cluster]) -> Cluster:
    if len(clusters) == 1:
        return clusters[0]
    return ask_pick_cluster(clusters=clusters)


def ask_pick_workspace(workspaces: List[Workspace]) -> Workspace:
    choices = [
        questionary.Choice(title=f"{w.name} ({w.fqn})", value=w) for w in workspaces
    ]
    return questionary.select("Pick a workspace", choices=choices).ask()


def maybe_ask_pick_workspace(workspaces: List[Workspace]) -> Workspace:
    if len(workspaces) == 1:
        return workspaces[0]
    return ask_pick_workspace(workspaces=workspaces)


def resolve_cluster_or_error(
    name_or_id: Optional[str] = None,
    ignore_context: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Cluster:
    if non_interactive:
        if ignore_context and not name_or_id:
            raise ValueError("cluster name or id cannot be null")

    clusters = resolve_clusters(
        client=client, name_or_id=name_or_id, ignore_context=ignore_context
    )

    if not clusters:
        if name_or_id:
            raise ValueError(f"No cluster found with name or id {name_or_id!r}")
        else:
            raise ValueError(f"No clusters found!")
    else:
        if non_interactive:
            if len(clusters) > 1:
                raise ValueError(
                    f"More than one cluster found with name or id {name_or_id!r}: {clusters!r}"
                )
            else:
                cluster = clusters[0]
        else:
            cluster = maybe_ask_pick_cluster(clusters=clusters)
    return cluster


def resolve_workspace_or_error(
    name_or_id: Optional[str] = None,
    cluster_name_or_id: Optional[Union[Cluster, str]] = None,
    ignore_context: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Tuple[Workspace, Cluster]:
    if non_interactive:
        if ignore_context and not name_or_id:
            raise ValueError("workspace name or id cannot be null")

    if isinstance(cluster_name_or_id, Cluster):
        cluster = cluster_name_or_id
    else:
        cluster = resolve_cluster_or_error(
            name_or_id=cluster_name_or_id,
            non_interactive=non_interactive,
            ignore_context=ignore_context,
            client=client,
        )

    console.print(PROMPT_USING_CLUSTER_CONTEXT.format(cluster.name))

    workspaces = resolve_workspaces(
        client=client,
        name_or_id=name_or_id,
        cluster_name_or_id=cluster,
        ignore_context=ignore_context,
    )
    if not workspaces:
        if name_or_id:
            raise ValueError(
                f"No workspace found with name or id {name_or_id!r} in cluster {cluster.name!r}"
            )
        else:
            raise ValueError(f"No workspaces found!")
    else:
        if non_interactive:
            if len(workspaces) > 1:
                raise ValueError(
                    f"More than one workspace found with name or id {name_or_id!r}: {workspaces!r}"
                )
            else:
                workspace = workspaces[0]
        else:
            workspace = maybe_ask_pick_workspace(workspaces=workspaces)
    return workspace, cluster



import importlib
import inspect
import os
import sys
import traceback

from servicefoundry.requirements.interceptor.communication import (
    Communication,
    Dependencies,
    Function,
    Invoke,
)
from servicefoundry.requirements.interceptor.module_interceptor import ModuleInterceptor


def main(file_name, communication):
    sys.path.insert(1, os.getcwd())
    module_interceptor = ModuleInterceptor()
    module = importlib.import_module(file_name.split(".")[0])
    communication.write_ack()
    while True:
        process_command(module, module_interceptor, communication)


def process_command(module, module_interceptor, communication):
    command = communication.read_input_msg()
    if isinstance(command, Invoke):
        try:
            method_to_call = getattr(module, command.function)
            res = method_to_call(**command.payload)
            communication.write_response(res)
        except Exception as e:
            communication.write_exception(str(e), traceback.format_exc())
    elif isinstance(command, Dependencies):
        communication.write_response(module_interceptor.get_dependencies())
    elif isinstance(command, Function):
        functions = [
            member[0]
            for member in inspect.getmembers(module, predicate=inspect.isfunction)
        ]
        communication.write_response(functions)
    else:
        communication.write_exception(f"Can only handle command of type {[Invoke]}")


if __name__ == "__main__":
    infile = os.fdopen(int(sys.argv[2]))
    outfile = os.fdopen(int(sys.argv[3]), "w", buffering=1)
    communication = Communication(infile, outfile)
    try:
        main(sys.argv[1], communication)
    except Exception as e:
        print("Closing Interceptor.")
        traceback.format_exc()
        communication.write_exception(str(e), traceback.format_exc())
        communication.close()



from typing import List

from docker.errors import APIError, BuildError

import docker
from servicefoundry.logger import logger

__all__ = [
    "push_docker_image",
    "docker_build",
    "pull_docker_image",
    "push_docker_image_with_latest_tag",
]


def _get_docker_client():
    try:
        return docker.from_env()
    except Exception as ex:
        raise Exception("Could not connect to Docker") from ex


# this is required since push does throw an error if it
# fails - so we have to parse the response logs to catch the error
# the other option is to run `docker login` as a subprocess but it's
# not recommended to provide password to subprocess
def catch_error_in_push(response: List[dict]):
    for line in response:
        if line.get("error") is not None:
            raise Exception(
                f'Failed to push to registry with message \'{line.get("error")}\''
            )


def build_docker_image(
    path: str, tag: str, platform: str, dockerfile: str, cache_from: List[str]
):
    logger.info("Starting docker build...")
    try:
        for result in _get_docker_client().api.build(
            decode=True,
            path=path,
            tag=tag,
            platform=platform,
            dockerfile=dockerfile,
            cache_from=cache_from,
        ):
            for (_, value) in result.items():
                if isinstance(value, str):
                    logger.info(value)
    except (BuildError, APIError) as ex:
        raise Exception("Error while building Docker image") from ex


def push_docker_image(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    client = _get_docker_client()
    auth_config = {"username": docker_login_username, "password": docker_login_password}

    logger.info(f"Pushing {image_uri}")
    response = client.images.push(
        repository=image_uri, auth_config=auth_config, decode=True, stream=True
    )
    catch_error_in_push(response=response)


def push_docker_image_with_latest_tag(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    client = _get_docker_client()
    auth_config = {"username": docker_login_username, "password": docker_login_password}

    repository_without_tag, _ = image_uri.rsplit(":", 1)
    image = client.images.get(image_uri)
    image.tag(repository=repository_without_tag, tag="latest")

    logger.info(f"Pushing {repository_without_tag}:latest")
    response = client.images.push(
        repository=repository_without_tag,
        tag="latest",
        auth_config=auth_config,
        decode=True,
        stream=True,
    )
    catch_error_in_push(response=response)


def pull_docker_image(
    image_uri: str,
    docker_login_username: str,
    docker_login_password: str,
):
    auth_config = {"username": docker_login_username, "password": docker_login_password}
    logger.info(f"Pulling cache image {image_uri}")
    _get_docker_client().images.pull(repository=image_uri, auth_config=auth_config)



from servicefoundry.v2.lib import (
    Application,
    BasicAuthCreds,
    Build,
    DockerFileBuild,
    FileMount,
    GitSource,
    HealthProbe,
    HttpProbe,
    HuggingfaceModelHub,
    Image,
    Job,
    LocalSource,
    Manual,
    ModelDeployment,
    Notebook,
    Param,
    Port,
    PythonBuild,
    RemoteSource,
    Resources,
    Schedule,
    Service,
    TruefoundryModelRegistry,
)



# generated by datamodel-codegen:
#   filename:  application.json
#   timestamp: 2023-01-03T06:27:37+00:00

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, confloat, conint, constr
from typing_extensions import Literal


class BasicAuthCreds(BaseModel):
    """
    +label=Username and password for service auth
    """

    type: Literal["basic_auth"]
    username: constr(regex=r"^[a-z0-9-@.]{1,64}$") = Field(
        ...,
        description="+label=Username for service auth\n+message=Upto 64 lower case alphanumeric character long",
    )
    password: constr(regex=r"^[a-zA-Z0-9.!@#$%^&*()_+-=]{1,64}$") = Field(
        ...,
        description="+label=Password for service auth\n+message=Password should not be more than 64 characters",
    )


class DockerFileBuild(BaseModel):
    """
        +docs=Describes that we are using a dockerfile to build our image
    +label=Docker File (I already have Docker File)
    +icon=fa-brands fa-docker:#0db7ed
    """

    type: constr(regex=r"dockerfile")
    dockerfile_path: str = Field(
        "./Dockerfile",
        description="+label=Path to Dockerfile\n+usage=The file path of the Dockerfile relative to project root path.",
    )
    build_context_path: str = Field(
        "./",
        description="+label=Path to build context\n+usage=Build context path for the Dockerfile relative to project root path.",
    )
    command: Optional[Union[str, List[str]]] = Field(
        None,
        description="+label=Entrypoint Override\n+usage=Override the command to run when the container starts\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )


class FileMount(BaseModel):
    mount_dir: str = Field(..., description="+usage=Dir at which data is to be mounted")
    data: Dict[str, str] = Field(
        ...,
        description="+usage=Data to be mounted, the key will be the filename, and the value will be the file content. files will be mounted under mount_dir",
    )


class GitSource(BaseModel):
    """
        +docs=Describes that we are using code stored in a git repository to build our image
    +label=Git Source
    +icon=fa-solid fa-code-branch:black
    """

    type: constr(regex=r"git")
    repo_url: constr(
        regex=r"(http(s?)://)(github.com|(.+@)*bitbucket.org|gitlab.com).*$"
    ) = Field(
        ...,
        description="+label=Repo URL\n+usage=The repository URL.\n+sort=1\n+message=Needs to be a valid Github, Bitbucket or Gitlab link",
    )
    ref: str = Field(
        ..., description="+label=Commit SHA\n+usage=The commit SHA.\n+sort=2"
    )
    branch_name: Optional[str] = Field(
        None,
        description="+label=Branch Name\n+usage=Selecting branch will select latest commit SHA of the branch.\n+sort=3\n+ignore",
    )


class HttpProbe(BaseModel):
    """
        +docs=Describes the Instructions for assessing container health by executing an HTTP GET request.
    To learn more you can go [here](https://docs.truefoundry.com/docs/add-health-checks-to-deployments)
    +label=Instructions for assessing container health by executing an HTTP GET request.
    """

    type: constr(regex=r"http")
    path: str = Field(
        ...,
        description="+usage=The endpoint, relative to the port, to which the HTTP GET request should be directed.",
    )
    port: conint(ge=0, le=65535) = Field(
        ...,
        description="+usage=The TCP socket within the container to which the HTTP GET request should be directed.",
    )
    host: Optional[str] = None
    scheme: str = "HTTP"


class HuggingfaceModelHub(BaseModel):
    """
        +docs=Huggingface model hub (deploy a model from HF model hub)
    +label=Huggingface model hub
    +usage=Deploy a model from HF model hub
    """

    type: constr(regex=r"hf-model-hub")
    model_id: constr(regex=r"[A-Za-z0-9_.-/]+") = Field(
        ...,
        description="+label=Model repository name\n+usage=The name of the model that you want to deploy. Example:- `t5-small`, `philschmid/bart-large-cnn-samsum`.\n+placeholder=philschmid/bart-large-cnn-samsum",
    )
    pipeline: Optional[constr(regex=r"[a-z-]+")] = Field(
        None,
        description="+label=Pipeline\n+usage=Pipeline to use for inference. If not set, we will try to infer the `task` name for the given model.\nExamples:- `summarization`, `text-generation`, `text-classification`, etc.\nYou can read more about HF pipelines here:- https://huggingface.co/docs/transformers/pipeline_tutorial\nYou can find a list of tasks here:- https://huggingface.co/docs/transformers/v4.24.0/en/main_classes/pipelines#transformers.pipeline.task\n+placeholder=summarization",
    )


class Image(BaseModel):
    """
        +docs=Describes that we are using a pre-built image stored in a Docker Image registry
    +label=Docker Image (Deploy an existing image)
    +icon=fa-brands fa-docker:#0db7ed
    """

    type: constr(regex=r"image")
    image_uri: str = Field(
        ...,
        description="+label=Image URI\n+usage=The image URI. Specify the name of the image and the tag.\nIf the image is in Dockerhub, you can skip registry-url (for e.g. `tensorflow/tensorflow`).\nYou can use an image from a private registry using Advanced fields\n+placeholder=registry-url/account/image:version (e.g. docker.io/tensorflow/tensorflow)",
    )
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )
    command: Optional[Union[str, List[str]]] = Field(
        None,
        description="+label=Entrypoint Override\n+usage=Override the command to run when container starts.\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )


class Restart(str, Enum):
    """
    +usage=When to restart the job
    """

    Never = "Never"
    OnFailure = "OnFailure"
    Always = "Always"


class LocalSource(BaseModel):
    """
        +docs=Describes that we are using code stored in a local developement environment to build our image
    +label=Local
    +icon=fa-folder:black
    """

    type: constr(regex=r"local")
    project_root_path: str = Field("./", description="+usage=Local project root path.")
    local_build: bool = Field(False, description="run docker build locally")


class Manual(BaseModel):
    """
        +docs=Describes that we are going to manually trigger our job.
    +label=Manual
    +usage=Trigger the job manually. [Docs](https://docs.truefoundry.com/docs/more-configurations)
    """

    type: constr(regex=r"manual")


class Param(BaseModel):
    name: constr(regex=r"^[a-z][a-z0-9-_]{0,30}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the param\n+message=name can contain lower case alphabets, digits, underscore (_) and hypen (-). It can be 32 characters long, should start with an alphabet, and should end with either an alphabet or digit.",
    )
    description: Optional[constr(regex=r"^.{1,127}$")] = Field(
        None,
        description="+usage=Description of param\n+message=description cannot be longer than 127 characters",
    )
    default: Optional[constr(regex=r"^.{1,127}$")] = Field(
        None,
        description="+usage=Default value or placeholder\n+message=default value cannot be longer than 127 characters",
    )


class Protocol(str, Enum):
    """
    +usage=Protocol for the port.
    """

    TCP = "TCP"
    UDP = "UDP"


class Port(BaseModel):
    """
    +docs=Describes the ports the service should be exposed to.
    """

    protocol: Protocol = Field("TCP", description="+usage=Protocol for the port.")
    port: conint(ge=0, le=65535) = Field(
        80, description="+usage=Port number to expose."
    )
    expose: bool = Field(True, description="+usage=Expose the port")
    host: Optional[
        constr(
            regex=r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"
        )
    ] = Field(None, description="+usage=Host")
    path: Optional[
        constr(regex=r"^(/([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9]))*/$")
    ] = Field(None, description="+usage=Path")
    auth: Optional[BasicAuthCreds] = None


class PythonBuild(BaseModel):
    """
        +docs=Describes that we are using python to build a container image with a specific python version and pip packages installed.
    +label=Python Buildpack (I don't have Dockerfile)
    +icon=fa-brands fa-python:#306998
    """

    type: constr(regex=r"tfy-python-buildpack")
    python_version: constr(regex=r"^\d+(\.\d+){1,2}$") = Field(
        "3.9",
        description="+label=Python version\n+usage=Python version to run your application.\n+message=Please enter a valid Python version",
    )
    build_context_path: str = Field(
        "./",
        description="+label=Path to build context\n+usage=Build path relative to project root path.",
    )
    requirements_path: Optional[str] = Field(
        None,
        description="`Path to build context`\n+label=Path to requirements\n+usage=Path to `requirements.txt` relative to\n`Path to build context`",
    )
    pip_packages: Optional[List[str]] = Field(
        None,
        description="+label=Pip packages to install\n+usage=Define pip package requirements.",
    )
    apt_packages: Optional[List[str]] = Field(
        None,
        description='+label=Debian packages to install via apt.\n+usage=Define Debian packages to install via apt.\nE.g. `["ffmpeg", "libsm6", "libxext6"]',
    )
    command: Union[str, List[str]] = Field(
        ...,
        description="Command will be set as the Entrypoint of the generated\nimage.\n+label=Entrypoint\n+usage=Command to run when the container starts.\nCommand will be set as the Entrypoint of the generated image.\nWhen deploying a Job, the command can be templatized by defining `params` and referencing them in command\nE.g. `python main.py --learning_rate {{learning_rate}}`",
    )


class RemoteSource(BaseModel):
    """
        +docs=Describes that we are using code stored in a remote respository to build our image
    +label=S3
    +icon=fa-brands fa-aws:black
    """

    type: constr(regex=r"remote")
    remote_uri: str = Field(
        ..., description="+docs=Remote repository URI\n+label=Remote URI"
    )


class Resources(BaseModel):
    """
        +docs=Describes the resource constraints for the application so that it can be deployed accordingly on the cluster
    To learn more you can go [here](https://docs.truefoundry.com/docs/resources)
    +icon=fa-microchip
    +label=Resources
    +usage=Set resource constraints for the application. [Docs](https://docs.truefoundry.com/docs/resources)
    """

    cpu_request: confloat(ge=0.001, le=16.0) = Field(
        0.2,
        description="+label=CPU Request\n+sort=1\n+usage=Requested CPU which determines the minimum cost incurred. The CPU usage can exceed the requested\namount, but not the value specified in the limit. 1 CPU means 1 CPU core. Fractional CPU can be requested\nlike `0.5` or `0.05`",
    )
    cpu_limit: confloat(ge=0.001, le=16.0) = Field(
        0.5,
        description="+label=CPU Limit\n+usage=CPU limit beyond which the usage cannot be exceeded. 1 CPU means 1 CPU core. Fractional CPU can be requested\nlike `0.5`. CPU limit should be >= cpu request.\n+sort=2",
    )
    memory_request: confloat(ge=1.0, le=32000.0) = Field(
        200,
        description="+label=Memory Request\n+usage=Requested memory which determines the minimum cost incurred. The unit of memory is in megabytes(MB).\nSo 1 means 1 MB and 2000 means 2GB.\n+sort=3",
    )
    memory_limit: confloat(ge=1.0, le=32000.0) = Field(
        500,
        description="+label=Memory Limit\n+usage=Memory limit after which the application will be killed with an OOM error. The unit of memory is\nin megabytes(MB). So 1 means 1 MB and 2000 means 2GB. MemoryLimit should be greater than memory request.\n+sort=4",
    )
    ephemeral_storage_request: float = Field(
        1000,
        description="+label=Storage Request\n+usage=Requested disk storage. The unit of memory is in megabytes(MB).\nThis is ephemeral storage and will be wiped out on pod restarts or eviction\n+sort=5",
    )
    ephemeral_storage_limit: float = Field(
        2000,
        description="+label=Storage Limit\n+usage=Disk storage limit. The unit of memory is in megabytes(MB). Exceeding this limit will result in eviction.\nIt should be greater than the request. This is ephemeral storage and will be wiped out on pod restarts or eviction\n+sort=6",
    )
    instance_family: Optional[List[str]] = Field(
        None,
        description="+label=Instance family\n+usage=Instance family of the underlying machine to use. Multiple instance families can be supplied.\nThe workload is guaranteed to be scheduled on one of them.",
    )


class ConcurrencyPolicy(str, Enum):
    """
        +usage=Choose whether to allow this job to run while another instance of the job is running, or to replace the currently running instance. Allow
    will enable multiple instances of this job to run. Forbid will keep the current instance of the job running and stop a new instance from being run.
    Replace will terminate any currently running instance of the job and start a new one.
    """

    Forbid = "Forbid"
    Allow = "Allow"
    Replace = "Replace"


class Schedule(BaseModel):
    """
        +docs=Describes that we are going to schedule our job to run at a schedule, making our job a cron job.
    +label=Schedule
    +usage=Run the job on a schedule. [Docs](https://docs.truefoundry.com/docs/deploy-a-cron-job)
    """

    type: constr(regex=r"scheduled")
    schedule: str = Field(
        ...,
        description="+docs=Specify the schedule for this job to be run periodically in cron format. [Learn more](https://docs.truefoundry.com/docs/deploy-a-cron-job)\n+usage=Specify the schedule for this job to be run periodically in cron format.\n```\n* * * * *\n| | | | |\n| | | | |___ day of week (0-6) (Sunday is 0)\n| | | |_____ month (1-12)\n| | |_______ day of month (1-31)\n| |_________ hour (0-23)\n|___________ minute (0-59)\n```",
    )
    concurrency_policy: ConcurrencyPolicy = Field(
        "Forbid",
        description="+usage=Choose whether to allow this job to run while another instance of the job is running, or to replace the currently running instance. Allow\nwill enable multiple instances of this job to run. Forbid will keep the current instance of the job running and stop a new instance from being run.\nReplace will terminate any currently running instance of the job and start a new one.",
    )


class TruefoundryModelRegistry(BaseModel):
    """
        +docs=Truefoundry model registry (deploy a model from TFY model registry)
    +label=Truefoundry model registry
    +usage=Deploy a model from TFY model registry
    """

    type: constr(regex=r"tfy-model-registry")
    model_version_fqn: constr(regex=r"model:[a-zA-Z0-9-/_]+:\d+") = Field(
        ...,
        description="+label=Model Version FQN\n+usage=The FQN of the model version that you want to deploy.",
    )


class Build(BaseModel):
    """
        +docs=Describes how we build our code into a Docker image.
    +label=Source Code (Build and deploy source code)
    +icon=fa-code
    """

    type: constr(regex=r"build")
    docker_registry: Optional[str] = Field(
        None,
        description="+docs=FQN of the container registry. You can the FQN of your desired container registry (or add one)\nin the  Integrations page[Integrations](https://app.truefoundry.tech/integrations?tab=docker-registry) page\n+label=Docker Registry\n+usage=FQN of the container registry. If you can't find your registry here,\nadd it through the [Integrations](/integrations?tab=docker-registry) page",
    )
    build_source: Union[RemoteSource, GitSource, LocalSource] = Field(
        ...,
        description="+docs=Source code location.\n+label=Fetch source code to build and deploy\n+icon=fa-code\n+sort=1",
    )
    build_spec: Union[DockerFileBuild, PythonBuild] = Field(
        ...,
        description="+docs=Instructions to build a container image out of the build source\n+label=Build using DockerFile or using Buildpack\n+icon=fa-wrench\n+sort=2",
    )


class HealthProbe(BaseModel):
    """
        +docs=Describes the configuration for the Health Probe's
    To learn more you can go [here](https://docs.truefoundry.com/docs/add-health-checks-to-deployments)
    +icon=fa-heart
    """

    config: HttpProbe
    initial_delay_seconds: conint(ge=0, le=36000) = Field(
        0,
        description="+usage=Number of seconds after the container is started before the first probe is initiated.",
    )
    period_seconds: conint(ge=0, le=36000) = Field(
        10, description="+usage=How often, in seconds, to execute the probe."
    )
    timeout_seconds: conint(ge=0, le=36000) = Field(
        1, description="+usage=Number of seconds after which the probe times out."
    )
    success_threshold: conint(ge=0, le=100) = Field(
        1,
        description="+usage=Minimum consecutive successes for the probe to be considered successful after having failed.",
    )
    failure_threshold: conint(ge=0, le=100) = Field(
        3,
        description="+usage=Number of consecutive failures required to determine the container is not alive (liveness probe) or not ready (readiness probe).",
    )


class Job(BaseModel):
    """
    +docs=Describes the configuration for the job
    """

    type: constr(regex=r"job")
    name: constr(regex=r"^[a-z][a-z0-9-]{1,23}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the job\n+sort=1\n+message=3 to 25 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    image: Union[Build, Image] = Field(
        ...,
        description="+docs=Specify whether you want to deploy a Docker image or build and deploy from source code\n+label=Deploy a Docker image or build and deploy from source code\n+icon=fa-solid fa-cloud-arrow-up:#21B6A8\n+sort=2",
    )
    resources: Optional[Resources] = None
    trigger: Union[Manual, Schedule] = Field(
        {"type": "manual"}, description="+docs=Specify the trigger\n+sort=3"
    )
    retries: conint(ge=0) = Field(
        1,
        description="+label=Retries\n+usage=Specify the maximum number of attempts to retry a job before it is marked as failed.\n+icon=fa-clone",
    )
    timeout: conint(ge=0, le=3600) = Field(
        1000,
        description="+label=Timeout\n+usage=Job timeout in seconds.\n+icon=fa-clone",
    )
    successful_jobs_history_limit: conint(ge=0, le=100) = Field(
        20,
        description="+usage=Number of successful invocation of this job to keep in history",
    )
    failed_jobs_history_limit: conint(ge=0, le=100) = Field(
        20,
        description="+usage=Number of failed invocation of this job to keep in history",
    )
    restart: Restart = Field("Never", description="+usage=When to restart the job")
    command: Optional[Union[str, List[str]]] = Field(
        None,
        description="+label=Commands\n+usage=Commands to execute to start the application. Overrides the `Entrypoint` command.\n+icon=fa-terminal\n+sort=6",
    )
    params: Optional[List[Param]] = Field(
        None,
        description="+label=Params for input\n+usage=Configure params and pass it to create different job runs\n+icon=fa-globe\n+sort=7",
    )
    env: Optional[Dict[str, str]] = Field(
        None,
        description="+label=Environment Variables\n+usage=Configure environment variables to be injected in the service. [Docs](https://docs.truefoundry.com/docs/env-variables)\n+icon=fa-globe\n+sort=8",
    )
    service_account: Optional[str] = Field(
        None, description="+usage=Service account that this workload should use"
    )
    file_mounts: Optional[List[FileMount]] = Field(
        None, description="+usage=Files to be mounted\n+ignore"
    )


class ModelDeployment(BaseModel):
    """
        +docs=Describes the configuration for the model deployment
    +usage=To learn more you can go [here](https://dash.readme.com/project/truefoundry/v0.1.1/docs/model-deployment-2)
    """

    type: constr(regex=r"model-deployment")
    name: constr(regex=r"^[a-z][a-z0-9-]{1,17}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the model deployment. This uniquely identifies this deployment in the workspace.\n> Name can only contain alphanumeric characters and '-' and can be atmost 20 characters long\n+sort=1\n+label=Model deployment name\n+message=3 to 19 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    model_source: Union[TruefoundryModelRegistry, HuggingfaceModelHub] = Field(
        ...,
        description="+docs=Specify the model to be deployed\n+sort=2\n+label=Model to deploy",
    )
    resources: Optional[Resources] = None
    grpc: bool = Field(False, description="+docs=Use grpc\n+sort=4\n+label=Use gRPC")


class Notebook(BaseModel):
    """
    +docs=Describes the configuration for the notebook
    """

    type: constr(regex=r"notebook")
    name: constr(regex=r"^[a-z][a-z0-9-]{1,23}[a-z0-9]$") = Field(
        ..., description="+usage=Name of the notebook"
    )
    resources: Optional[Resources] = None


class Service(BaseModel):
    """
    +docs=Describes the configuration for the service
    """

    type: constr(regex=r"service")
    name: constr(regex=r"^[a-z][a-z0-9-]{1,23}[a-z0-9]$") = Field(
        ...,
        description="+usage=Name of the service. This uniquely identifies this service in the workspace.\n> Name can only contain alphanumeric characters and '-' and can be atmost 25 characters long\n+sort=1\n+message=3 to 25 lower case characters long alphanumeric word, may contain - in between, cannot start with a number",
    )
    image: Union[Build, Image] = Field(
        ...,
        description="+docs=Specify whether you want to deploy a Docker image or build and deploy from source code\n+label=Deploy a Docker image or build and deploy from source code\n+icon=fa-solid fa-cloud-arrow-up:#21B6A8\n+sort=2",
    )
    resources: Optional[Resources] = None
    replicas: conint(ge=1, le=5) = Field(
        1,
        description="+label=Replicas\n+usage=Number of service instances/replicas you want to run.\n+icon=fa-clone\n+sort=3",
    )
    command: Optional[Union[str, List[str]]] = Field(
        None,
        description="+label=Commands\n+usage=Commands to execute to start the application. Overrides the `Entrypoint` command.\n+icon=fa-terminal",
    )
    env: Optional[Dict[str, str]] = Field(
        None,
        description="+label=Environment Variables\n+usage=Configure environment variables to be injected in the service. [Docs](https://docs.truefoundry.com/docs/env-variables)\n+icon=fa-globe\n+sort=6",
    )
    ports: List[Port] = Field(
        ...,
        description="+docs=Specify the ports you want the service to be exposed to\n+label=Ports (Ports to route customer traffic to)\n+icon=fa-plug\n+sort=4",
    )
    liveness_probe: Optional[HealthProbe] = None
    readiness_probe: Optional[HealthProbe] = None
    service_account: Optional[str] = Field(
        None, description="+usage=Service account that this workload should use"
    )
    file_mounts: Optional[List[FileMount]] = Field(
        None, description="+usage=Files to be mounted\n+ignore"
    )


class Application(BaseModel):
    __root__: Union[Service, Job, Notebook, ModelDeployment]



import inspect
import os
import subprocess
import sys
from pathlib import Path

from servicefoundry.requirements.interceptor import main
from servicefoundry.requirements.interceptor.communication import (
    Ack,
    CloseException,
    ErrorResponse,
    InvokeException,
)
from servicefoundry.requirements.interceptor.main import Communication


class Interceptor:
    def __init__(self, predict_file):
        path = Path(predict_file)
        if not path.exists():
            raise RuntimeError(f"{predict_file} not exist.")
        self.predict_file = predict_file

        if predict_file.endswith(".py"):
            self.module_name = predict_file[0:-3].replace(os.sep, ".")
        else:
            raise RuntimeError(f"{predict_file} doesn't end with py")

        (self.read1, self.write1) = os.pipe()  # for parent -> child writes
        (self.read2, self.write2) = os.pipe()  # for child -> parent writes
        command = [
            sys.executable,
            inspect.getfile(main),
            predict_file,
            str(self.read1),
            str(self.write2),
        ]
        self.process = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            pass_fds=(self.read1, self.write2),
        )
        infile = os.fdopen(self.read2)
        outfile = os.fdopen(self.write1, "w", buffering=1)
        self.communication = Communication(infile, outfile)
        msg = self.communication.read_input_msg()
        if isinstance(msg, Ack):
            return
        if isinstance(msg, ErrorResponse):
            raise RuntimeError(f"Failed to load predictor. Cause by: {msg.message}")

    def _read_response(self):
        try:
            return self.communication.read_response()
        except InvokeException as e:
            print(e.traceback)
            raise e
        except CloseException as e:
            self.close()
            print(e.traceback)
            raise e

    def invoke(self, function_name, **kwargs):
        self.communication.write_invoke(function_name, **kwargs)
        return self._read_response()

    def get_dependencies(self):
        self.communication.write_dependencies()
        return self._read_response()

    def get_functions(self):
        self.communication.write_function()
        return self._read_response()

    def close(self):
        self.process.terminate()
        for line in self.process.stdout.readlines():
            print(line)
        self.communication.close()

    @classmethod
    def create(cls, predict_file):
        return cls(predict_file)



SFY = "sfy"

# TODO: probably create another `rich_messages.py` and apply all formatting there
PROMPT_LOGIN_SUCCESSFUL = f"""[green bold]Login Successful![/]"""
PROMPT_LOGIN_INFO = (
    f"""[yellow]You are logged in as {{username!r}} with email {{email!r}}[/]"""
)
PROMPT_LOGOUT_SUCCESSFUL = f"""[green bold]Logged Out![/]"""
PROMPT_POST_LOGIN = ""
PROMPT_ALREADY_LOGGED_OUT = f"""[yellow]You are already logged out[/]"""
PROMPT_SETTING_CLUSTER_CONTEXT = f"""[yellow]Setting cluster {{!r}} as default[/]"""
PROMPT_SETTING_WORKSPACE_CONTEXT = f"""[yellow]Setting workspace {{!r}} as default[/]"""
PROMPT_UNSETTING_WORKSPACE_CONTEXT = (
    f"""[yellow]Removing workspace {{!r}} as default[/]"""
)
PROMPT_USING_CLUSTER_CONTEXT = f"""[yellow]Using cluster {{!r}}[/]"""
PROMPT_USING_WORKSPACE_CONTEXT = f"""[yellow]Using workspace {{!r}}[/]"""
PROMPT_CREATING_NEW_WORKSPACE = f"""[yellow]Creating a new workspace {{!r}}[/]"""
PROMPT_DELETING_WORKSPACE = f"""[yellow]Deleting workspace {{!r}}[/]"""
PROMPT_DELETED_WORKSPACE = f"""[green]Deleted workspace {{!r}}[/]"""
PROMPT_DELETING_SERVICE = f"""[yellow]Deleting service {{!r}}[/]"""
PROMPT_DELETED_SERVICE = f"""[green]Deleted service {{!r}}[/]"""
PROMPT_NO_WORKSPACES = f"""[yellow]No workspaces found. You can create one with [bold]{SFY} create workspace[/][/]"""
PROMPT_NO_SERVICES = f"""[yellow]No services found. You can create one with [bold]{SFY} init[/] and then
deploy it using [bold]{SFY} deploy[/][/]"""
PROMPT_NO_DEPLOYMENTS = f"""[yellow]No deployments found. You can create one with [bold]{SFY} deploy .[/] from within
your service folder"""
PROMPT_NO_DEPLOYMENTS = f"""[yellow]No deployments found."""
PROMPT_LOGIN_POST_TENANT_CHANGE = f"""[yellow]Your session needs to be refreshed as a new tenant was configured.
Please login with [bold]{SFY} login[/]"""



import importlib
import sys
from importlib.abc import MetaPathFinder

import pkg_resources


class FinderInterceptor(MetaPathFinder):
    def __init__(self):
        super().__init__()
        self.modules = set()

    def find_spec(self, fullname, path, target=None):
        if path is None:
            self.modules.add(fullname)
        return None

    def get_modules(self):
        return self.modules


def get_version_from_attr(module_name):
    try:
        module = importlib.import_module(module_name)
        return module.__version__
    except Exception:
        return None


class ModuleInterceptor:
    def __init__(self):
        self.interceptor = FinderInterceptor()
        sys.meta_path.insert(0, self.interceptor)

    def get_dependencies(self):
        dependencies = {}
        intercepted_modules = self.interceptor.get_modules()
        for module in pkg_resources.working_set:
            module_name = module.key
            if module_name in intercepted_modules:
                dependencies[module_name] = module.version
                # version = get_version_from_attr(module_name)
                # if version is None:
                #     dependencies[module_name] = module.version
                # else:
                #     dependencies[module_name] = version
        return dependencies



import os.path
import re

from pkg_resources import Requirement

MANAGED_MESSAGE = """
# ServicefoundryManaged
# Below this line dependencies was added by servicefoundry.
""".splitlines()


class SfPackage:
    def __init__(self, package_name, package_version, is_installed=False):
        self.package_name = package_name
        self.package_version = package_version
        self.is_installed = is_installed

    def update_version(self, version):
        self.is_installed = True
        self.package_version = version

    def to_str(self):
        if self.package_version:
            return f"{self.package_name}=={self.package_version}"
        else:
            return f"{self.package_name}"


class PythonRequirements:
    def __init__(self, old_requirements_txt: str):
        self.old_user_lines = []
        self.old_packages = set()
        self.final_packages = {}
        source_lines = old_requirements_txt.splitlines()
        self._load(source_lines)

    def _load(self, source_lines):
        is_managed_block = False
        for line in source_lines:
            if re.match(r"^\s*#\s*ServicefoundryManaged\s*$", line):
                is_managed_block = True
            if not is_managed_block:
                self.old_user_lines.append(line)
            sline = line.strip()
            if sline.strip() != "" and re.match(r"^\s*#", line) is None:
                requirement = Requirement(line)
                package_name = requirement.key.lower()
                if not is_managed_block:
                    self.old_packages.add(package_name)
                else:
                    try:
                        version = requirement.specs[0][1]
                    except IndexError:
                        version = None
                    self.final_packages[package_name] = SfPackage(package_name, version)

    def update_requirements_txt(self, new_dependencies: dict):
        for package_name, package_version in new_dependencies.items():
            # To prevent servicefoundry==0.0.0 lines
            if package_version != "0.0.0":
                if package_name not in self.old_packages:

                    if "+" in package_version:
                        package_version = package_version.split("+")[0]

                    if package_name in self.final_packages:
                        self.final_packages[package_name].update_version(
                            package_version
                        )
                    else:
                        self.final_packages[package_name] = SfPackage(
                            package_name, package_version, is_installed=True
                        )

    def get_requirements_txt(self):
        # Generate lines.
        ret_lines = []
        for line in self.old_user_lines:
            ret_lines.append(line)
        ret_lines.extend(MANAGED_MESSAGE)
        for package_name, package in sorted(self.final_packages.items()):
            ret_lines.append(package.to_str())
        return os.linesep.join(ret_lines)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['servicefoundry',
 'servicefoundry.auto_gen',
 'servicefoundry.builder',
 'servicefoundry.builder.builders',
 'servicefoundry.builder.builders.tfy_python_buildpack',
 'servicefoundry.cli',
 'servicefoundry.cli.commands',
 'servicefoundry.cli.io',
 'servicefoundry.core',
 'servicefoundry.core.notebook',
 'servicefoundry.core.notebook.io',
 'servicefoundry.function_service',
 'servicefoundry.function_service.deployment_examples.class_deployment',
 'servicefoundry.function_service.deployment_examples.function_deployment',
 'servicefoundry.function_service.deployment_examples.model_composition',
 'servicefoundry.function_service.deployment_examples.one_service_calling_another',
 'servicefoundry.function_service.remote',
 'servicefoundry.internal',
 'servicefoundry.io',
 'servicefoundry.lib',
 'servicefoundry.lib.auth',
 'servicefoundry.lib.clients',
 'servicefoundry.lib.config',
 'servicefoundry.lib.dao',
 'servicefoundry.lib.infra',
 'servicefoundry.lib.model',
 'servicefoundry.requirements',
 'servicefoundry.requirements.interceptor',
 'servicefoundry.v2',
 'servicefoundry.v2.examples.job_deployment',
 'servicefoundry.v2.examples.model_deployment.hf',
 'servicefoundry.v2.examples.model_deployment.mlfoundry',
 'servicefoundry.v2.examples.service_deployment',
 'servicefoundry.v2.lib']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.27,<4.0.0',
 'Mako>=1.1.6,<2.0.0',
 'PyJWT>=2.3.0,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'chevron>=0.14.0,<0.15.0',
 'click>=8.0.4,<9.0.0',
 'cookiecutter>=2.1.1,<3.0.0',
 'docker>=6.0.1,<7.0.0',
 'fastapi>=0.78.0,<0.79.0',
 'filelock>=3.8.0,<4.0.0',
 'gitignorefile>=1.1.2,<2.0.0',
 'importlib-metadata>=4.2,<5.0',
 'importlib-resources>=5.2.0,<6.0.0',
 'packaging>=21.3,<22.0',
 'pydantic>=1.9.1,<2.0.0',
 'pygments>=2.12.0,<3.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'python-socketio[client]>=5.5.2,<6.0.0',
 'questionary>=1.10.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'rich-click>=1.2.1,<2.0.0',
 'rich>=12.0.0,<13.0.0',
 'typing-extensions>=4.3.0,<5.0.0',
 'uvicorn>=0.18.2,<0.19.0']

extras_require = \
{'notebook': ['ipywidgets>=7.6.0,<8.0.0', 'ipython>=7.10.0,<8.0.0']}

entry_points = \
{'console_scripts': ['servicefoundry = servicefoundry.cli.__main__:main',
                     'sfy = servicefoundry.cli.__main__:main']}

setup_kwargs = {
    'name': 'servicefoundry',
    'version': '0.6.6rc1',
    'description': 'Generate deployed services from code',
    'long_description': "# ServiceFoundry\n\nServiceFoundry is a client library that allows you to containerize and deploy your Machine Learning model (or other\nservices) to a managed Kubernetes Cluster. This also generates a Grafana cluster with complete visibility of your\nService Health, System Logs, and Kubernetes Workspace.\n\nIt is available both as a command-line-interface and via APIs that allow you to deploy directly from your Jupyter\nNotebook.\n\nYou can access the health of your service, monitoring links and deployed endpoints by logging on to TrueFoundry's\ndashboard.\n\n# Installation\n\n```\npip install -U servicefoundry\n```\n\n# Documentation\n\nhttps://docs.truefoundry.com/servicefoundry/\n",
    'author': 'Abhishek Choudhary',
    'author_email': 'abhichoudhary06@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/innoavator/servicefoundry',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import logging

import rich_click as click

from servicefoundry.cli.config import CliConfig
from servicefoundry.cli.const import COMMAND_CLS, ENABLE_CLUSTER_COMMANDS, GROUP_CLS
from servicefoundry.cli.display_util import print_json, print_obj
from servicefoundry.cli.util import handle_exception_wrapper
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.dao import workspace as workspace_lib
from servicefoundry.lib.model.entity import Cluster, Workspace

logger = logging.getLogger(__name__)

# TODO (chiragjn): --json should disable all non json console prints


@click.group(name="get", cls=GROUP_CLS)
def get_command():
    # TODO (chiragjn): Figure out a way to update supported resources based on ENABLE_* flags
    """
    Get servicefoundry resources

    \b
    Supported resources:
    - Workspace
    - Service
    - Deployment
    """
    pass


@click.command(name="cluster", cls=COMMAND_CLS, help="Get Cluster metadata")
@click.argument("cluster_id")
@handle_exception_wrapper
def get_cluster(cluster_id):
    tfs_client = ServiceFoundryServiceClient()
    cluster = tfs_client.get_cluster(cluster_id)
    if CliConfig.get("json"):
        print_json(data=cluster)
    else:
        print_obj("Cluster", cluster, columns=Cluster.get_display_columns)


@click.command(name="workspace", cls=COMMAND_CLS, help="Get Workspace metadata")
@click.argument("name", type=click.STRING)
@click.option(
    "-c",
    "--cluster",
    type=click.STRING,
    default=None,
    help="cluster to find this workspace in",
)
@click.option("--non-interactive", is_flag=True, default=False)
@handle_exception_wrapper
def get_workspace(name, cluster, non_interactive):
    workspace = workspace_lib.get_workspace(
        name_or_id=name,
        cluster_name_or_id=cluster,
        non_interactive=non_interactive,
    )
    if CliConfig.get("json"):
        print_json(data=workspace.to_dict())
    else:
        print_obj(
            "Workspace", workspace.to_dict(), columns=Workspace.get_display_columns
        )


def get_get_command():
    get_command.add_command(get_workspace)
    if ENABLE_CLUSTER_COMMANDS:
        get_command.add_command(get_cluster)
    return get_command



from typing import Any, Dict, Optional

from servicefoundry.cli.console import console
from servicefoundry.lib.clients.service_foundry_client import (
    ServiceFoundryServiceClient,
)
from servicefoundry.lib.messages import (
    PROMPT_CREATING_NEW_WORKSPACE,
    PROMPT_DELETED_WORKSPACE,
    PROMPT_DELETING_WORKSPACE,
    PROMPT_UNSETTING_WORKSPACE_CONTEXT,
    PROMPT_USING_CLUSTER_CONTEXT,
)
from servicefoundry.lib.model.entity import Workspace
from servicefoundry.lib.util import (
    all_workspaces,
    resolve_cluster_or_error,
    resolve_workspace_or_error,
    resolve_workspaces,
)


def create_workspace(
    name: str,
    cluster_name_or_id: Optional[str] = None,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient()
    cluster = resolve_cluster_or_error(
        name_or_id=cluster_name_or_id, non_interactive=non_interactive, client=client
    )
    console.print(PROMPT_USING_CLUSTER_CONTEXT.format(cluster.name))
    with console.status(PROMPT_CREATING_NEW_WORKSPACE.format(name), spinner="dots"):
        response = client.create_workspace(cluster_id=cluster.id, name=name)
        workspace = Workspace.from_dict(response["workspace"])

    return workspace


def get_workspace(
    name_or_id: str,
    cluster_name_or_id: Optional[str] = None,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient()
    workspace, _ = resolve_workspace_or_error(
        name_or_id=name_or_id,
        cluster_name_or_id=cluster_name_or_id,
        non_interactive=non_interactive,
        client=client,
    )
    return workspace


def get_workspace_by_fqn(
    workspace_fqn: str,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Workspace:
    client = client or ServiceFoundryServiceClient()
    workspaces = client.get_workspace_by_fqn(workspace_fqn=workspace_fqn)
    if len(workspaces) == 0:
        raise ValueError(f"Workspace with FQN {workspace_fqn!r} does not exist.")
    workspace = workspaces[0]
    return Workspace.from_dict(workspace)


def list_workspaces(
    cluster_name_or_id: Optional[str] = None,
    all_: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
):
    client = client or ServiceFoundryServiceClient()
    if all_:
        workspaces = all_workspaces(client=client)
    else:
        cluster = resolve_cluster_or_error(
            name_or_id=cluster_name_or_id,
            non_interactive=non_interactive,
            client=client,
        )
        console.print(PROMPT_USING_CLUSTER_CONTEXT.format(cluster.name))
        workspaces = resolve_workspaces(
            name_or_id=None,
            cluster_name_or_id=cluster,
            ignore_context=True,
            client=client,
        )
    return workspaces


def delete_workspace(
    name_or_id: str,
    cluster_name_or_id: Optional[str] = None,
    force: bool = False,
    non_interactive: bool = True,
    client: Optional[ServiceFoundryServiceClient] = None,
) -> Dict[str, Any]:
    client = client or ServiceFoundryServiceClient()
    workspace = get_workspace(
        name_or_id=name_or_id,
        cluster_name_or_id=cluster_name_or_id,
        non_interactive=non_interactive,
        client=client,
    )
    with console.status(
        PROMPT_DELETING_WORKSPACE.format(workspace.name), spinner="dots"
    ):
        response = client.remove_workspace(workspace.id, force=force)
        ctx_workspace = client.session.get_workspace()
        if ctx_workspace:
            ctx_workspace = Workspace.from_dict(ctx_workspace)
            if ctx_workspace.id == workspace.id:
                client.session.set_workspace(None)
                console.print(
                    PROMPT_UNSETTING_WORKSPACE_CONTEXT.format(ctx_workspace.name)
                )
                client.session.save_session()
    console.print(PROMPT_DELETED_WORKSPACE.format(workspace.name))
    return response
